import re
import os
import subprocess
import json
import shutil
from typing import TYPE_CHECKING, List, Optional, Tuple, Dict, Any

from . import segment_map
from .verilog_preprocessor import get_illegal_trailing_text_pos
from ..source_ref import DirectSourceRef, FileSourceRef

if TYPE_CHECKING:
    from typing import Set
    from ..compiler import RDLEnvironment

class PerlPreprocessor:

    def __init__(self, env: 'RDLEnvironment', path: str, search_paths: List[str], incl_ref: Optional[segment_map.IncludeRef]=None):
        self.env = env
        self.path = path
        self.search_paths = search_paths
        self.incl_ref = incl_ref

        with open(path, 'r', newline='', encoding='utf_8') as f:
            self.text = f.read()

        self.included_files = set() # type: Set[str]

    #---------------------------------------------------------------------------
    def preprocess(self) -> Tuple[str, segment_map.SegmentMap]:
        """
        Run preprocessor on a top-level file.

        Performs the following preprocess steps:

        - Expand `include directives
        - Perl Preprocessor

        Returns
        -------
        tuple
            (preprocessed_text, SegmentMap)
        """
        tokens = self.tokenize()
        pl_segments, has_perl_tags = self.get_perl_segments(tokens)

        # Generate flattened output
        str_parts = []
        smap = segment_map.SegmentMap()
        offset = 0

        if has_perl_tags:
            # Needs to be processed through perl interpreter
            emit_list = self.run_perl_miniscript(pl_segments)

            for entry in emit_list:
                if entry['type'] == "ref":
                    pl_seg = pl_segments[entry['ref']]
                    emit_text = pl_seg.get_text()

                    ref_seg = segment_map.UnalteredSegment(
                        offset, offset + len(emit_text) - 1,
                        pl_seg.start, pl_seg.end, pl_seg.file_pp.path,
                        pl_seg.file_pp.incl_ref
                    )
                    offset += len(emit_text)
                    smap.segments.append(ref_seg)
                    str_parts.append(emit_text)

                elif entry['type'] == "text":
                    pl_seg = pl_segments[entry['ref']]
                    emit_text = entry['text']

                    macro_seg = segment_map.MacroSegment(
                        offset, offset + len(emit_text) - 1,
                        pl_seg.start, pl_seg.end, pl_seg.file_pp.path,
                        pl_seg.file_pp.incl_ref
                    )
                    offset += len(emit_text)
                    smap.segments.append(macro_seg)
                    str_parts.append(emit_text)
        else:
            # OK to bypass perl interpreter
            for pl_seg in pl_segments:
                emit_text = pl_seg.get_text()
                map_seg = segment_map.UnalteredSegment(
                    offset, offset + len(emit_text) - 1,
                    pl_seg.start, pl_seg.end, pl_seg.file_pp.path,
                    pl_seg.file_pp.incl_ref
                )
                offset += len(emit_text)
                smap.segments.append(map_seg)
                str_parts.append(emit_text)

        return ("".join(str_parts), smap)

    #---------------------------------------------------------------------------
    def tokenize(self) -> List[Tuple[str, int, int]]:
        """
        Tokenize the input text

        Scans for instances of perl tags and include directives.
        Tokenization skips line and block comments.

        Returns
        -------
        list
            List of tuples: (typ, start, end)

            Where:

            - typ is "perl" or "incl"
            - start/end mark the first/last char offset of the token
        """
        tokens = []
        token_spec = [
            ('mlc', r'/\*.*?\*/'),
            ('slc', r'//.*?$'),
            ('perl', r'<%.*?%>'),
            ('incl', r'`include'),
        ]
        tok_regex = '|'.join('(?P<%s>%s)' % pair for pair in token_spec)
        for m in re.finditer(tok_regex, self.text, re.DOTALL | re.MULTILINE):
            if m.lastgroup in ("incl", "perl"):
                tokens.append((m.lastgroup, m.start(0), m.end(0)-1))
        return tokens

    #---------------------------------------------------------------------------
    def parse_include(self, start: int) -> Tuple[int, str]:
        """
        Extract include from text based on start position of token

        Returns
        -------
        (end, incl_path)
            - end: last char in include
            - incl_path: Resolved path to include
        """
        # Seek back to start of line
        i = start
        while i:
            if self.text[i] == '\n':
                i += 1
                break
            i -= 1
        line_start = i

        # check that there is no unexpected text before the include
        if not (self.text[line_start:start] == "" or self.text[line_start:start].isspace()):
            self.env.msg.fatal(
                "Unexpected text before include",
                DirectSourceRef(self.path, line_start, start-1)
            )

        # Capture include contents
        inc_regex = re.compile(r'`include\s+("([^\r\n]+)"|<([^\r\n]+)>)')
        m_inc = inc_regex.match(self.text, start)
        if m_inc is None:
            self.env.msg.fatal(
                "Invalid usage of include directive",
                DirectSourceRef(self.path, start, start+7)
            )
        incl_path_raw = m_inc.group(2) or m_inc.group(3)
        end = m_inc.end(0)-1
        path_start = m_inc.start(1)

        # Check that only comments follow
        cruft_start, cruft_end = get_illegal_trailing_text_pos(self.text, end+1)
        if cruft_start is not None and cruft_end is not None:
            self.env.msg.fatal(
                "Unexpected text after include",
                DirectSourceRef(self.path, cruft_start, cruft_end)
            )

        # Resolve include path.
        if os.path.isabs(incl_path_raw):
            incl_path = incl_path_raw
        else:
            # Search include paths first.
            for search_path in self.search_paths:
                incl_path = os.path.join(search_path, incl_path_raw)
                if os.path.isfile(incl_path):
                    # found match!
                    break
            else:
                # Otherwise, assume it is relative to the current file
                incl_path = os.path.join(os.path.dirname(self.path), incl_path_raw)
        if not os.path.isfile(incl_path):
            self.env.msg.fatal(
                "Could not find '%s' in include search paths" % incl_path_raw,
                DirectSourceRef(self.path, path_start, end)
            )

        # Check if path has already been referenced before
        incl_ref = self.incl_ref
        while incl_ref:
            if os.path.samefile(incl_path, incl_ref.path):
                self.env.msg.fatal(
                    "Include of '%s' results in a circular reference" % incl_path_raw,
                    DirectSourceRef(self.path, path_start, end)
                )
            incl_ref = incl_ref.parent

        return(end, incl_path)

    #---------------------------------------------------------------------------
    def get_perl_segments(self, tokens: List[Tuple[str, int, int]]) -> Tuple[List['PPPSegment'], bool]:
        """
        Build a list of perl preprocessor segments:
            PPPUnalteredSegment
            PPPPerlSegment
            PPPMacroSegment
        returns:
            (pl_segments, has_perl_tags)
        """
        pl_segments = [] # type: List[PPPSegment]
        has_perl_tags = False
        pos = 0

        for typ, start, end in tokens:

            # Capture any leading text
            if start != pos:
                pl_seg = PPPUnalteredSegment(self, pos, start-1) # type: PPPSegment
                pl_segments.append(pl_seg)

            if typ == "incl":
                # Got an `include ...

                # Extract the path and actual end position
                end, incl_path = self.parse_include(start)

                # create a reference that captures the location where the include occurred
                incl_ref = segment_map.IncludeRef(start, end, self.path, self.incl_ref)


                # Recurse and extract perl segments from included file
                incl_file_pp = PerlPreprocessor(self.env, incl_path, self.search_paths, incl_ref)
                incl_tokens = incl_file_pp.tokenize()
                incl_pl_segments, incl_has_pl_tags = incl_file_pp.get_perl_segments(incl_tokens)

                self.included_files.add(incl_path)
                self.included_files.update(incl_file_pp.included_files)
                pl_segments.extend(incl_pl_segments)
                has_perl_tags = has_perl_tags or incl_has_pl_tags

            else:
                # Got a Perl tag <% ... %>
                if self.text[start+2] == "=":
                    pl_seg = PPPMacroSegment(self, start, end)
                else:
                    pl_seg = PPPPerlSegment(self, start, end)
                pl_segments.append(pl_seg)
                has_perl_tags = True

            pos = end+1

        # Capture any trailing text
        text_len = len(self.text)
        if text_len > pos:
            pl_seg = PPPUnalteredSegment(self, pos, text_len-1)
            pl_segments.append(pl_seg)

        return (pl_segments, has_perl_tags)

    #---------------------------------------------------------------------------
    def run_perl_miniscript(self, segments: List['PPPSegment']) -> List[Dict[str, Any]]:
        """
        Generates and runs a perl miniscript that derives the text that will be
        emitted from the preprocessor

        returns the resulting emit list
        """

        # Check if perl is installed
        if shutil.which("perl") is None:
            self.env.msg.fatal(
                "Input contains Perl preprocessor tags, but an installation of Perl could not be found"
            )

        # Generate minimal perl script that captures activities described in the source file
        lines = []
        for i, pp_seg in enumerate(segments):
            if isinstance(pp_seg, PPPUnalteredSegment):
                # Text outside preprocessor tags that should remain unaltered
                # Insert command to emit reference to this text segment
                lines.append("rdlppp_utils::emit_ref(%d);" % i)

            elif isinstance(pp_seg, PPPPerlSegment):
                # Perl code snippet. Insert directly
                lines.append(pp_seg.get_text())

            elif isinstance(pp_seg, PPPMacroSegment):
                # Preprocessor macro print tag
                # Insert command to store resulting text
                var = pp_seg.get_text()

                # Check for any illegal characters
                if re.match(r'[\s;]', var):
                    self.env.msg.fatal(
                        "Invalid text found in Perl macro expansion",
                        DirectSourceRef(self.path, pp_seg.start, pp_seg.end)
                    )

                lines.append("rdlppp_utils::emit_text(%d, %s);" % (i, var))
        miniscript = '\n'.join(lines)

        # Run miniscript
        runner_path = os.path.join(os.path.dirname(__file__), "ppp_runner.pl")
        result = subprocess.run(
            ["perl", runner_path, ",".join(self.env.perl_safe_opcodes)],
            input=miniscript.encode("utf-8"),
            stdout=subprocess.PIPE, stderr=subprocess.PIPE,
            timeout=5, check=False
        )
        if result.returncode:
            self.env.msg.fatal(
                "Encountered a Perl syntax error while executing embedded Perl preprocessor commands:\n"
                + result.stderr.decode("utf-8"),
                # TODO: Fix useless context somehow
                FileSourceRef(self.path)
            )

        # miniscript returns the emit list in JSON format. Convert it
        emit_list = json.loads(result.stdout.decode('utf-8'))

        return emit_list

#-------------------------------------------------------------------------------

class PPPSegment:
    def __init__(self, file_pp: PerlPreprocessor, start: int, end: int):
        self.file_pp = file_pp
        self.start = start
        self.end = end

    def get_text(self) -> str:
        return self.file_pp.text[self.start:self.end+1]

class PPPUnalteredSegment(PPPSegment):
    pass

class PPPPerlSegment(PPPSegment):
    def get_text(self) -> str:
        return self.file_pp.text[self.start+2:self.end-1]

class PPPMacroSegment(PPPSegment):
    def get_text(self) -> str:
        return self.file_pp.text[self.start+3:self.end-1]



# Generated from SystemRDL.g4 by ANTLR 4.13.1
from antlr4 import *
if "." in __name__:
    from .SystemRDLParser import SystemRDLParser
else:
    from SystemRDLParser import SystemRDLParser

# This class defines a complete generic visitor for a parse tree produced by SystemRDLParser.

class SystemRDLVisitor(ParseTreeVisitor):

    # Visit a parse tree produced by SystemRDLParser#root.
    def visitRoot(self, ctx:SystemRDLParser.RootContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by SystemRDLParser#eval_expr_root.
    def visitEval_expr_root(self, ctx:SystemRDLParser.Eval_expr_rootContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by SystemRDLParser#root_elem.
    def visitRoot_elem(self, ctx:SystemRDLParser.Root_elemContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by SystemRDLParser#component_def.
    def visitComponent_def(self, ctx:SystemRDLParser.Component_defContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by SystemRDLParser#explicit_component_inst.
    def visitExplicit_component_inst(self, ctx:SystemRDLParser.Explicit_component_instContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by SystemRDLParser#component_inst_alias.
    def visitComponent_inst_alias(self, ctx:SystemRDLParser.Component_inst_aliasContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by SystemRDLParser#component_named_def.
    def visitComponent_named_def(self, ctx:SystemRDLParser.Component_named_defContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by SystemRDLParser#component_anon_def.
    def visitComponent_anon_def(self, ctx:SystemRDLParser.Component_anon_defContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by SystemRDLParser#component_body.
    def visitComponent_body(self, ctx:SystemRDLParser.Component_bodyContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by SystemRDLParser#component_body_elem.
    def visitComponent_body_elem(self, ctx:SystemRDLParser.Component_body_elemContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by SystemRDLParser#component_insts.
    def visitComponent_insts(self, ctx:SystemRDLParser.Component_instsContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by SystemRDLParser#component_inst.
    def visitComponent_inst(self, ctx:SystemRDLParser.Component_instContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by SystemRDLParser#field_inst_reset.
    def visitField_inst_reset(self, ctx:SystemRDLParser.Field_inst_resetContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by SystemRDLParser#inst_addr_fixed.
    def visitInst_addr_fixed(self, ctx:SystemRDLParser.Inst_addr_fixedContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by SystemRDLParser#inst_addr_stride.
    def visitInst_addr_stride(self, ctx:SystemRDLParser.Inst_addr_strideContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by SystemRDLParser#inst_addr_align.
    def visitInst_addr_align(self, ctx:SystemRDLParser.Inst_addr_alignContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by SystemRDLParser#component_inst_type.
    def visitComponent_inst_type(self, ctx:SystemRDLParser.Component_inst_typeContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by SystemRDLParser#component_type.
    def visitComponent_type(self, ctx:SystemRDLParser.Component_typeContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by SystemRDLParser#component_type_primary.
    def visitComponent_type_primary(self, ctx:SystemRDLParser.Component_type_primaryContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by SystemRDLParser#param_def.
    def visitParam_def(self, ctx:SystemRDLParser.Param_defContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by SystemRDLParser#param_def_elem.
    def visitParam_def_elem(self, ctx:SystemRDLParser.Param_def_elemContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by SystemRDLParser#param_inst.
    def visitParam_inst(self, ctx:SystemRDLParser.Param_instContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by SystemRDLParser#param_assignment.
    def visitParam_assignment(self, ctx:SystemRDLParser.Param_assignmentContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by SystemRDLParser#BinaryExpr.
    def visitBinaryExpr(self, ctx:SystemRDLParser.BinaryExprContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by SystemRDLParser#UnaryExpr.
    def visitUnaryExpr(self, ctx:SystemRDLParser.UnaryExprContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by SystemRDLParser#NOP.
    def visitNOP(self, ctx:SystemRDLParser.NOPContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by SystemRDLParser#TernaryExpr.
    def visitTernaryExpr(self, ctx:SystemRDLParser.TernaryExprContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by SystemRDLParser#expr_primary.
    def visitExpr_primary(self, ctx:SystemRDLParser.Expr_primaryContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by SystemRDLParser#concatenate.
    def visitConcatenate(self, ctx:SystemRDLParser.ConcatenateContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by SystemRDLParser#replicate.
    def visitReplicate(self, ctx:SystemRDLParser.ReplicateContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by SystemRDLParser#paren_expr.
    def visitParen_expr(self, ctx:SystemRDLParser.Paren_exprContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by SystemRDLParser#CastType.
    def visitCastType(self, ctx:SystemRDLParser.CastTypeContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by SystemRDLParser#CastWidth.
    def visitCastWidth(self, ctx:SystemRDLParser.CastWidthContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by SystemRDLParser#cast_width_expr.
    def visitCast_width_expr(self, ctx:SystemRDLParser.Cast_width_exprContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by SystemRDLParser#range_suffix.
    def visitRange_suffix(self, ctx:SystemRDLParser.Range_suffixContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by SystemRDLParser#array_suffix.
    def visitArray_suffix(self, ctx:SystemRDLParser.Array_suffixContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by SystemRDLParser#array_type_suffix.
    def visitArray_type_suffix(self, ctx:SystemRDLParser.Array_type_suffixContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by SystemRDLParser#data_type.
    def visitData_type(self, ctx:SystemRDLParser.Data_typeContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by SystemRDLParser#basic_data_type.
    def visitBasic_data_type(self, ctx:SystemRDLParser.Basic_data_typeContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by SystemRDLParser#literal.
    def visitLiteral(self, ctx:SystemRDLParser.LiteralContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by SystemRDLParser#NumberInt.
    def visitNumberInt(self, ctx:SystemRDLParser.NumberIntContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by SystemRDLParser#NumberHex.
    def visitNumberHex(self, ctx:SystemRDLParser.NumberHexContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by SystemRDLParser#NumberVerilog.
    def visitNumberVerilog(self, ctx:SystemRDLParser.NumberVerilogContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by SystemRDLParser#string_literal.
    def visitString_literal(self, ctx:SystemRDLParser.String_literalContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by SystemRDLParser#boolean_literal.
    def visitBoolean_literal(self, ctx:SystemRDLParser.Boolean_literalContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by SystemRDLParser#array_literal.
    def visitArray_literal(self, ctx:SystemRDLParser.Array_literalContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by SystemRDLParser#struct_literal.
    def visitStruct_literal(self, ctx:SystemRDLParser.Struct_literalContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by SystemRDLParser#struct_kv.
    def visitStruct_kv(self, ctx:SystemRDLParser.Struct_kvContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by SystemRDLParser#enum_literal.
    def visitEnum_literal(self, ctx:SystemRDLParser.Enum_literalContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by SystemRDLParser#accesstype_literal.
    def visitAccesstype_literal(self, ctx:SystemRDLParser.Accesstype_literalContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by SystemRDLParser#onreadtype_literal.
    def visitOnreadtype_literal(self, ctx:SystemRDLParser.Onreadtype_literalContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by SystemRDLParser#onwritetype_literal.
    def visitOnwritetype_literal(self, ctx:SystemRDLParser.Onwritetype_literalContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by SystemRDLParser#addressingtype_literal.
    def visitAddressingtype_literal(self, ctx:SystemRDLParser.Addressingtype_literalContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by SystemRDLParser#precedencetype_literal.
    def visitPrecedencetype_literal(self, ctx:SystemRDLParser.Precedencetype_literalContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by SystemRDLParser#instance_ref.
    def visitInstance_ref(self, ctx:SystemRDLParser.Instance_refContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by SystemRDLParser#instance_ref_element.
    def visitInstance_ref_element(self, ctx:SystemRDLParser.Instance_ref_elementContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by SystemRDLParser#prop_ref.
    def visitProp_ref(self, ctx:SystemRDLParser.Prop_refContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by SystemRDLParser#local_property_assignment.
    def visitLocal_property_assignment(self, ctx:SystemRDLParser.Local_property_assignmentContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by SystemRDLParser#dynamic_property_assignment.
    def visitDynamic_property_assignment(self, ctx:SystemRDLParser.Dynamic_property_assignmentContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by SystemRDLParser#normal_prop_assign.
    def visitNormal_prop_assign(self, ctx:SystemRDLParser.Normal_prop_assignContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by SystemRDLParser#encode_prop_assign.
    def visitEncode_prop_assign(self, ctx:SystemRDLParser.Encode_prop_assignContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by SystemRDLParser#prop_mod_assign.
    def visitProp_mod_assign(self, ctx:SystemRDLParser.Prop_mod_assignContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by SystemRDLParser#prop_assignment_rhs.
    def visitProp_assignment_rhs(self, ctx:SystemRDLParser.Prop_assignment_rhsContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by SystemRDLParser#prop_keyword.
    def visitProp_keyword(self, ctx:SystemRDLParser.Prop_keywordContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by SystemRDLParser#prop_mod.
    def visitProp_mod(self, ctx:SystemRDLParser.Prop_modContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by SystemRDLParser#udp_def.
    def visitUdp_def(self, ctx:SystemRDLParser.Udp_defContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by SystemRDLParser#udp_attr.
    def visitUdp_attr(self, ctx:SystemRDLParser.Udp_attrContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by SystemRDLParser#udp_type.
    def visitUdp_type(self, ctx:SystemRDLParser.Udp_typeContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by SystemRDLParser#udp_data_type.
    def visitUdp_data_type(self, ctx:SystemRDLParser.Udp_data_typeContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by SystemRDLParser#udp_usage.
    def visitUdp_usage(self, ctx:SystemRDLParser.Udp_usageContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by SystemRDLParser#udp_comp_type.
    def visitUdp_comp_type(self, ctx:SystemRDLParser.Udp_comp_typeContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by SystemRDLParser#udp_default.
    def visitUdp_default(self, ctx:SystemRDLParser.Udp_defaultContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by SystemRDLParser#udp_constraint.
    def visitUdp_constraint(self, ctx:SystemRDLParser.Udp_constraintContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by SystemRDLParser#enum_def.
    def visitEnum_def(self, ctx:SystemRDLParser.Enum_defContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by SystemRDLParser#enum_entry.
    def visitEnum_entry(self, ctx:SystemRDLParser.Enum_entryContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by SystemRDLParser#enum_prop_assign.
    def visitEnum_prop_assign(self, ctx:SystemRDLParser.Enum_prop_assignContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by SystemRDLParser#struct_def.
    def visitStruct_def(self, ctx:SystemRDLParser.Struct_defContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by SystemRDLParser#struct_elem.
    def visitStruct_elem(self, ctx:SystemRDLParser.Struct_elemContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by SystemRDLParser#struct_type.
    def visitStruct_type(self, ctx:SystemRDLParser.Struct_typeContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by SystemRDLParser#constraint_def.
    def visitConstraint_def(self, ctx:SystemRDLParser.Constraint_defContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by SystemRDLParser#constraint_named_def.
    def visitConstraint_named_def(self, ctx:SystemRDLParser.Constraint_named_defContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by SystemRDLParser#constraint_anon_def.
    def visitConstraint_anon_def(self, ctx:SystemRDLParser.Constraint_anon_defContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by SystemRDLParser#constraint_body.
    def visitConstraint_body(self, ctx:SystemRDLParser.Constraint_bodyContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by SystemRDLParser#constraint_body_elem.
    def visitConstraint_body_elem(self, ctx:SystemRDLParser.Constraint_body_elemContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by SystemRDLParser#constraint_insts.
    def visitConstraint_insts(self, ctx:SystemRDLParser.Constraint_instsContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by SystemRDLParser#constr_relational.
    def visitConstr_relational(self, ctx:SystemRDLParser.Constr_relationalContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by SystemRDLParser#constr_prop_assign.
    def visitConstr_prop_assign(self, ctx:SystemRDLParser.Constr_prop_assignContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by SystemRDLParser#constr_inside_values.
    def visitConstr_inside_values(self, ctx:SystemRDLParser.Constr_inside_valuesContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by SystemRDLParser#constr_inside_enum.
    def visitConstr_inside_enum(self, ctx:SystemRDLParser.Constr_inside_enumContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by SystemRDLParser#constr_lhs.
    def visitConstr_lhs(self, ctx:SystemRDLParser.Constr_lhsContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by SystemRDLParser#constr_inside_value.
    def visitConstr_inside_value(self, ctx:SystemRDLParser.Constr_inside_valueContext):
        return self.visitChildren(ctx)



del SystemRDLParser


from typing import TYPE_CHECKING, List, Set, Tuple, Optional, Dict

from .perl_preprocessor import PerlPreprocessor
from .verilog_preprocessor import VerilogPreprocessor
from .stream import PreprocessedInputStream
#from . import segment_map

if TYPE_CHECKING:
    from ..compiler import RDLEnvironment

def preprocess_file(env: 'RDLEnvironment', path: str, search_paths: List[str], defines: Optional[Dict[str,str]]=None) -> Tuple[PreprocessedInputStream, Set[str]]:
    # Run file through Perl preprocessor
    ppp = PerlPreprocessor(env, path, search_paths)
    preprocessed_text, seg_map = ppp.preprocess()
    included_files = ppp.included_files

    # ... then through the Verilog preprocessor
    vpp = VerilogPreprocessor(env, preprocessed_text, seg_map, defines=defines)
    preprocessed_text, seg_map = vpp.preprocess()

    #segment_map.print_segment_debug(preprocessed_text, seg_map)

    # Encapsulate into an Antlr-like input stream object
    input_stream = PreprocessedInputStream(preprocessed_text, seg_map)
    return input_stream, included_files




#: Check if a field that implements storage, or is a runtime constant, is
#: missing its reset value.
MISSING_RESET = 1<<0

#: Check if a field's bit offset is not explicitly specified.
#:
#: Some organizations may want to enforce explicit assignment of bit offsets to
#: avoid unexpected field packing.
IMPLICIT_FIELD_POS = 1<<1

#: Check if a component's address offset is not explicitly assigned.
#:
#: Some organizations may want to enforce explicit assignment of addresses to
#: avoid unintended address map changes.
IMPLICIT_ADDR = 1<<2

#: Check if an instance array's address stride is not a power of two.
STRIDE_NOT_POW2 = 1<<3

#: Enforce that all addressable components are aligned based on their size.
#: Alignment is determined by the component's size rounded up to the next power
#: of two.
#:
#: Strict self-alignment may be desireable since it can simplify address decode
#: logic for hierarchical designs.
#:
#: This rule is a superset of ``STRIDE_NOT_POW2``.
STRICT_SELF_ALIGN = 1<<4

#: Check if an array of registers uses a stride that is not equal to the
#: register's width.
#:
#: Many export formats are unable to natively represent register arrays that are
#: not tightly packed. (IP-XACT, UVM Virtual registers, C arrays, etc..)
SPARSE_REG_STRIDE = 1<<5

#-------------------------------------------------------------------------------
#: Enable all warnings.
ALL = (
    MISSING_RESET
    | IMPLICIT_FIELD_POS
    | IMPLICIT_ADDR
    | STRIDE_NOT_POW2
    | STRICT_SELF_ALIGN
    | SPARSE_REG_STRIDE
)



from typing import Tuple, List, Optional, Type, TYPE_CHECKING, Any
import copy

from ..node import Node, AddressableNode
from .. import component as comp
from .user_struct import UserStruct, is_user_struct

if TYPE_CHECKING:
    from ..source_ref import SourceRefBase
    from ..compiler import RDLEnvironment


RefElement = Tuple[str, List[int], Optional['SourceRefBase']]

class ComponentRef:
    """
    Container for hierarchical component instance references.
    This is used internally to store information about the reference
    When a user requests the reference value, it is resolved into a Node object
    """

    def __init__(self, ref_root: 'comp.Component', ref_elements: List[RefElement]):
        # Handle to the component definition where ref_elements is relative to
        # This is the original_def, and NOT the actual instance
        self.ref_root = ref_root

        # List of hierarchical reference element tuples that make up the path
        # to the reference.
        # Path is relative to the instance of ref_root
        # Each tuple in the list represents a segment of the path:
        # [
        #   ( <ID string> , [ <Index int> , ... ], SourceRefBase ),
        # ]
        self.ref_elements = ref_elements

    def build_node_ref(self, assignee_node: Node) -> Node:
        """
        Resolves the component reference into a Node object
        """
        current_node = assignee_node
        # Traverse up from assignee until ref_root is reached
        while True:
            if current_node is None:
                raise RuntimeError("Upwards traverse to ref_root failed")
            if current_node.inst.original_def is self.ref_root:
                break
            current_node = current_node.parent

        for inst_name, idx_list, name_src_ref in self.ref_elements:
            # find instance
            current_node = current_node.get_child_by_name(inst_name)


            # Check if indexes are valid
            if idx_list:
                # Reference contains one or more suffixes
                # Validation during compilation would have already enforced that
                # references are sane.
                # Safe to expect this to be an AddressableNode
                assert isinstance(current_node, AddressableNode)

                for i, idx in enumerate(idx_list):
                    if idx >= current_node.array_dimensions[i]:
                        current_node.env.msg.fatal(
                            "Array index out of range. Expected 0-%d, got %d."
                            % (current_node.array_dimensions[i]-1, idx),
                            name_src_ref
                        )

                # Assign indexes if appropriate
                if current_node.is_array:
                    current_node.current_idx = idx_list

        return current_node


class PropertyReference:
    """
    Base class for all property references used in RHS of an expression.

    The PropertyReference object represents the expression's reference target.
    Details of the reference can be determined using its ``node`` and ``name``
    variables.

    For example, the following property assignment:

    .. code-block:: systemrdl

        reg {
            ...
            fieldX->next = fieldY->intr;
        } my_reg;

    ... can be queried as follows:

    .. code-block:: python

        fieldX = my_reg.get_child_by_name("fieldX")
        fieldY = my_reg.get_child_by_name("fieldY")

        next_prop = fieldX.get_property('next')
        print(next_prop.node == fieldY) # prints: True
        print(next_prop.name) # prints: "intr"

    """
    allowed_inst_type = None # type: Type[comp.Component]

    def __init__(self, src_ref: 'SourceRefBase', env: 'RDLEnvironment', comp_ref: ComponentRef):
        self.env = env
        self.src_ref = src_ref
        self._comp_ref = comp_ref

        #: Node object that represents the component instance from which the
        #: property is being referenced.
        self.node = None # type: Node

    def __repr__(self) -> str:
        return f"<{self.__class__.__qualname__} {self.node.get_path()}->{self.name} at {id(self):#x}>"

    def __eq__(self, other: object) -> bool:
        """
        Property reference equality checks determine whether the other object
        represents the same reference (node represents the same position in the
        register model's hierarchy and the property referenced is the same)
        """
        if not isinstance(other, PropertyReference):
            return NotImplemented
        if self.name != other.name:
            return False
        return self.node.get_path() == other.node.get_path()

    @property
    def name(self) -> str:
        """
        Name of the property being referenced
        """
        return self.get_name()

    @classmethod
    def get_name(cls) -> str:
        return cls.__name__.replace("PropRef_", "")

    def _resolve_node(self, assignee_node: Node) -> None:
        self.node = self._comp_ref.build_node_ref(assignee_node)

    def get_resolved_ref(self, assignee_node: Node) -> 'PropertyReference':
        # Return a copy of this PropertyReference, but with a resolved
        # node reference.
        # Copying is essential to prevent quirk found in bug #164
        new_prop_ref = copy.copy(self)
        new_prop_ref._resolve_node(assignee_node)
        return new_prop_ref

    def _validate(self) -> None:
        pass

    @property
    def width(self) -> Optional[int]:
        """
        Get the equivalent width of the property reference.

        Returns None if reference doesn't have a specifically defined width

        .. versionadded:: 1.21
        """
        raise NotImplementedError


class RefType:
    """
    This class it not a true RDL type, but rather an object that represents
    all possible RHS references that can be made. This represents all possible
    component references, as well as RHS property references.

    This is used primarily when pre-registering a UDP that has the generic
    'ref' datatype.

    .. versionadded:: 1.25
    """

    expanded = (
        comp.Field,
        comp.Signal,
        comp.Reg,
        comp.Regfile,
        comp.Addrmap,
        comp.Mem,
        PropertyReference,
        # TODO: eventually add constraint references
    )


def resolve_node_refs_in_array(assignee_node: Node, array: List[Any]) -> List[Any]:
    """
    Recursively converts any ComponentRef within the array to proper Node objects
    """
    new_array = []
    changed = False
    for value in array:
        if isinstance(value, ComponentRef):
            value = value.build_node_ref(assignee_node)
            changed = True
        elif isinstance(value, PropertyReference):
            value = value.get_resolved_ref(assignee_node)
            changed = True
        elif is_user_struct(type(value)):
            new_value = resolve_node_refs_in_struct(assignee_node, value)
            if new_value is not value:
                changed = True
            value = new_value
        new_array.append(value)

    if changed:
        return new_array
    return array

def resolve_node_refs_in_struct(assignee_node: Node, struct: UserStruct) -> UserStruct:
    """
    Recursively converts any ComponentRef within the array to proper Node objects
    """
    new_values = {}
    changed = False
    for member_name, value in struct.members.items():
        if isinstance(value, ComponentRef):
            value = value.build_node_ref(assignee_node)
            changed = True
        elif isinstance(value, PropertyReference):
            value = value.get_resolved_ref(assignee_node)
            changed = True
        elif is_user_struct(type(value)):
            new_svalue = resolve_node_refs_in_struct(assignee_node, value)
            if new_svalue is not value:
                changed = True
            value = new_svalue
        elif isinstance(value, list):
            new_avalue = resolve_node_refs_in_array(assignee_node, value)
            if new_avalue is not value:
                changed = True
            value = new_avalue
        new_values[member_name] = value

    if changed:
        cls = type(struct)
        new_struct = cls(new_values)
        return new_struct
    return struct



from typing import TYPE_CHECKING, Optional, Type

from .ast_node import ASTNode
from .conditional import is_castable

from ..core.helpers import truncate_int

if TYPE_CHECKING:
    from ..compiler import RDLEnvironment
    from ..source_ref import SourceRefBase

    OptionalSourceRef = Optional[SourceRefBase]

# Integer unary operators:
#   +  -  ~
# Normal expression context rules
class _UnaryIntExpr(ASTNode):
    def __init__(self, env: 'RDLEnvironment', src_ref: 'OptionalSourceRef', n: ASTNode):
        super().__init__(env, src_ref)
        self.n = n

    def predict_type(self) -> Type[int]:
        op_type = self.n.predict_type()
        if not is_castable(op_type, int):
            self.msg.fatal(
                "Operand of expression is not a compatible numeric type",
                self.src_ref
            )
        return int

    def get_min_eval_width(self) -> int:
        return self.n.get_min_eval_width()

class UnaryPlus(_UnaryIntExpr):
    def get_value(self, eval_width: Optional[int]=None) -> int:
        if eval_width is None:
            eval_width = self.get_min_eval_width()
        n = int(self.n.get_value(eval_width))
        return truncate_int(n, eval_width)

class UnaryMinus(_UnaryIntExpr):
    def get_value(self, eval_width: Optional[int]=None) -> int:
        if eval_width is None:
            eval_width = self.get_min_eval_width()
        n = int(self.n.get_value(eval_width))
        return truncate_int(-n, eval_width)

class BitwiseInvert(_UnaryIntExpr):
    def get_value(self, eval_width: Optional[int]=None) -> int:
        if eval_width is None:
            eval_width = self.get_min_eval_width()
        n = int(self.n.get_value(eval_width))
        return truncate_int(~n, eval_width)



from typing import TYPE_CHECKING, Optional, Type

from .ast_node import ASTNode
from .conditional import is_castable

from ..core.helpers import truncate_int

if TYPE_CHECKING:
    from ..compiler import RDLEnvironment
    from ..source_ref import SourceRefBase

    OptionalSourceRef = Optional[SourceRefBase]

# Exponent & shift operators:
#   **  <<  >>
# Righthand operand is self-determined
class _ExpShiftExpr(ASTNode):
    def __init__(self, env: 'RDLEnvironment', src_ref: 'OptionalSourceRef', l: ASTNode, r: ASTNode):
        super().__init__(env, src_ref)
        self.l = l
        self.r = r

    def predict_type(self) -> Type[int]:
        l_type = self.l.predict_type()
        r_type = self.r.predict_type()
        if not is_castable(l_type, int):
            self.msg.fatal(
                "Left operand of expression is not a compatible numeric type",
                self.src_ref
            )
        if not is_castable(r_type, int):
            self.msg.fatal(
                "Right operand of expression is not a compatible numeric type",
                self.src_ref
            )
        return int

    def get_min_eval_width(self) -> int:
        # Righthand op has no influence in evaluation context
        return self.l.get_min_eval_width()

class Exponent(_ExpShiftExpr):
    def get_value(self, eval_width: Optional[int]=None) -> int:
        if eval_width is None:
            eval_width = self.l.get_min_eval_width()
        # Right operand is self-determined
        l = int(self.l.get_value(eval_width))
        r = int(self.r.get_value())
        return truncate_int(int(l ** r), eval_width)

class LShift(_ExpShiftExpr):
    def get_value(self, eval_width: Optional[int]=None) -> int:
        if eval_width is None:
            eval_width = self.l.get_min_eval_width()
        # Right operand is self-determined
        l = int(self.l.get_value(eval_width))
        r = int(self.r.get_value())
        return truncate_int(l << r, eval_width)

class RShift(_ExpShiftExpr):
    def get_value(self, eval_width: Optional[int]=None) -> int:
        if eval_width is None:
            eval_width = self.l.get_min_eval_width()
        # Right operand is self-determined
        l = int(self.l.get_value(eval_width))
        r = int(self.r.get_value())
        return truncate_int(l >> r, eval_width)



from typing import TYPE_CHECKING, Dict, List, Union, Type, Tuple

from .parameter import Parameter

from .. import component as comp

if TYPE_CHECKING:
    from ..compiler import RDLEnvironment
    from ..source_ref import SourceRefBase
    from .. import rdltypes
    from ..ast import ASTNode

TypeNSRef = Union[comp.Component, Type['rdltypes.UserEnum'], Type['rdltypes.UserStruct']]
TypeNSEntry = TypeNSRef
TypeNSScope = Dict[str, TypeNSEntry]

ElementNSRef = Union[comp.Component, Parameter]
ElementNSEntry = Tuple[ElementNSRef, comp.Component]
ElementNSScope = Dict[str, ElementNSEntry]

DefaultNSRef = Union['ASTNode', bool, 'rdltypes.InterruptType']
DefaultNSEntry = Tuple['SourceRefBase', DefaultNSRef]
DefaultNSScope = Dict[str, DefaultNSEntry]

class NamespaceRegistry():

    def __init__(self, env: 'RDLEnvironment'):
        self.env = env
        self.msg = env.msg

        self.type_ns_stack = [{}] # type: List[TypeNSScope]
        self.element_ns_stack = [{}] # type: List[ElementNSScope]
        self.default_property_ns_stack = [{}] # type: List[DefaultNSScope]

        # Control if Parameter objects are visible from parent scopes
        self.parent_parameters_visible = True

    def register_type(self, name: str, ref: TypeNSRef, src_ref: 'SourceRefBase') -> None:
        if name in self.type_ns_stack[-1]:
            self.msg.fatal(
                "Multiple declarations of type '%s'" % name,
                src_ref
            )
        self.type_ns_stack[-1][name] = ref

    def register_element(self, name: str, ref: ElementNSRef, parent_comp_def: comp.Component, src_ref: 'SourceRefBase') -> None:
        if name in self.element_ns_stack[-1]:
            self.msg.fatal(
                "Multiple declarations of instance '%s'" % name,
                src_ref
            )
        self.element_ns_stack[-1][name] = (ref, parent_comp_def)

    def register_default_property(self, name: str, ref: DefaultNSRef, src_ref: 'SourceRefBase', overwrite_ok: bool=False) -> None:
        if not overwrite_ok:
            if name in self.default_property_ns_stack[-1]:
                self.msg.fatal(
                    "Default property '%s' was already assigned in this scope" % name,
                    src_ref
                )

        self.default_property_ns_stack[-1][name] = (src_ref, ref)

    def lookup_type(self, name: str) -> TypeNSEntry:
        for scope in reversed(self.type_ns_stack):
            if name in scope:
                return scope[name]
        return None

    def lookup_element(self, name: str) -> ElementNSEntry:
        """
        Look up the element (component instance, or parameter) visible in the
        current scope.

        Returns a tuple:
            (element, parent_def)

        Where 'element' is the Component or Parameter being fetched
        and 'parent_def' is the component definition that encloses it.
        """
        for i, scope in enumerate(reversed(self.element_ns_stack)):
            if name in scope:
                el, parent_def = scope[name]
                if i == 0:
                    # Return anything from local namespace
                    return (el, parent_def)
                elif isinstance(el, comp.Signal):
                    # Signals are allowed to be found in parent namespaces
                    return (el, parent_def)
                elif self.parent_parameters_visible and isinstance(el, Parameter):
                    # Parameters are allowed to be found in parent namespaces,
                    # except in some contexts
                    return (el, parent_def)
        return (None, None)

    def get_default_properties(self, comp_type):
        # type: (Type[comp.Component]) -> DefaultNSScope
        """
        Returns a flattened dictionary of all default property assignments
        visible in the current scope that apply to the current component type.
        """
        # Flatten out all the default assignments that apply to the current scope
        # This does not include any default assignments made within the current
        # scope, so exclude those.
        props = {}
        for scope in self.default_property_ns_stack[:-1]:
            props.update(scope)

        # filter out properties that are not relevant
        prop_names = list(props.keys())
        for prop_name in prop_names:
            rule = self.env.property_rules.lookup_property(prop_name)
            if rule is None:
                self.msg.fatal(
                    "Unrecognized property '%s'" % prop_name,
                    props[prop_name][0]
                )
            if comp_type not in rule.bindable_to:
                del props[prop_name]

        return props

    def enter_scope(self) -> None:
        self.type_ns_stack.append({})
        self.element_ns_stack.append({})
        self.default_property_ns_stack.append({})

    def exit_scope(self) -> None:
        self.type_ns_stack.pop()
        self.element_ns_stack.pop()
        self.default_property_ns_stack.pop()



from typing import TYPE_CHECKING, Optional, Any, List, Tuple, Dict, Type
from copy import deepcopy

from .. import rdltypes
from .. import component as comp

from .ast_node import ASTNode
from .conditional import is_castable

if TYPE_CHECKING:
    from ..compiler import RDLEnvironment
    from ..core.parameter import Parameter
    from ..source_ref import SourceRefBase
    from rdltypes.typing import PreElabRDLType, RDLValue

    OptionalSourceRef = Optional[SourceRefBase]

class ParameterRef(ASTNode):
    def __init__(self, env: 'RDLEnvironment', src_ref: 'OptionalSourceRef', param: 'Parameter'):
        super().__init__(env, src_ref)
        self.param = param

    def predict_type(self) -> 'PreElabRDLType':
        return self.param.param_type

    def get_min_eval_width(self) -> int:
        if self.param.expr is None:
            self.msg.fatal(
                "Value for parameter '%s' was never assigned" % self.param.name,
                self.src_ref
            )
        return self.param.expr.get_min_eval_width()

    def get_value(self, eval_width: Optional[int]=None) -> Any:
        if self.param.expr is None:
            self.msg.fatal(
                "Value for parameter '%s' was never assigned" % self.param.name,
                self.src_ref
            )
        return self.param.expr.get_value(eval_width)


class ArrayIndex(ASTNode):
    def __init__(self, env: 'RDLEnvironment', src_ref: 'OptionalSourceRef', array: ASTNode, index: ASTNode):
        super().__init__(env, src_ref)
        self.array = array
        self.index = index

    def predict_type(self) -> 'PreElabRDLType':
        if not is_castable(self.index.predict_type(), int):
            self.msg.fatal(
                "Array index is not a compatible numeric type",
                self.index.src_ref
            )

        array_type = self.array.predict_type()
        if not isinstance(array_type, rdltypes.ArrayedType):
            self.msg.fatal(
                "Cannot index non-array type",
                self.array.src_ref
            )
        assert isinstance(array_type, rdltypes.ArrayedType)

        return array_type.element_type

    def get_min_eval_width(self) -> int:
        # TODO: Need to actually reach in and get eval width of array element
        return 64

    def get_value(self, eval_width: Optional[int]=None) -> Any:
        index = self.index.get_value()
        array = self.array.get_value()
        if index >= len(array):
            self.msg.fatal(
                "Array index '%d' is out of range" % index,
                self.src_ref
            )
        return array[index]


class MemberRef(ASTNode):
    def __init__(self, env: 'RDLEnvironment', src_ref: 'OptionalSourceRef', struct: ASTNode, member_name: str):
        super().__init__(env, src_ref)
        self.struct = struct
        self.member_name = member_name

    def predict_type(self): # type: ignore
        struct_type = self.struct.predict_type()

        if not rdltypes.is_user_struct(struct_type):
            self.msg.fatal(
                "Cannot reference member of non-struct type",
                self.struct.src_ref
            )

        if self.member_name not in struct_type._members:
            self.msg.fatal(
                "'%s' is not a valid member of struct type '%s'"
                % (self.member_name, struct_type.__name__),
                self.src_ref
            )

        return struct_type._members[self.member_name]

    def get_min_eval_width(self) -> int:
        # TODO: Need to actually reach in and get eval width of struct member
        return 64

    def get_value(self, eval_width: Optional[int]=None) -> Any:
        struct = self.struct.get_value()
        return struct._values[self.member_name]


RefElementsType = List[
    Tuple[str, List[ASTNode], 'OptionalSourceRef']
]
class InstRef(ASTNode):
    def __init__(self, env: 'RDLEnvironment', ref_root: comp.Component, ref_elements: RefElementsType):
        super().__init__(env, None) # single src_ref doesn't make sense for InstRef

        # Handle to the component definition where ref_elements is relative to
        # This is the original_def, and NOT the actual instance
        # When deepcopying, this is copied by reference in order to preserve
        # the original def object
        self.ref_root = ref_root

        # List of hierarchical reference element tuples that make up the path
        # to the reference.
        # Path is relative to ref_inst
        # Each tuple in the list represents a segment of the path:
        # [
        #   ( str , [ <Index ASTNode> , ... ] , SourceRefBase),
        #   ( str , [ <Index ASTNode> , ... ] , SourceRefBase)
        # ]
        self.ref_elements = ref_elements

    def __deepcopy__(self, memo: Dict[int, Any]) -> 'InstRef':
        """
        Copy any Source Ref by ref within the ref_elements list when deepcopying
        """
        copy_by_ref = ["env", "msg", "ref_root"]
        cls = self.__class__
        result = cls.__new__(cls)
        memo[id(self)] = result
        for k, v in self.__dict__.items():
            if k in copy_by_ref:
                setattr(result, k, v)
            elif k == "ref_elements":
                # Manually deepcopy the ref_elements list
                new_ref_elements = []
                for src_name, src_array_suffixes, src_src_ref in v:
                    new_array_suffixes = deepcopy(src_array_suffixes, memo)
                    new_ref_elements.append((src_name, new_array_suffixes, src_src_ref))
                setattr(result, k, new_ref_elements)
            else:
                setattr(result, k, deepcopy(v, memo))
        return result

    def predict_type(self) -> Type[comp.Component]:
        """
        Traverse the ref_elements path and determine the component type being
        referenced.
        Also do some checks on the array indexes
        """
        current_comp = self.ref_root
        for name, array_suffixes, name_src_ref in self.ref_elements:

            # find instance
            current_comp = current_comp.get_child_by_name(name)
            if current_comp is None:
                # Not found!
                self.msg.fatal(
                    "Could not resolve hierarchical reference to '%s'" % name,
                    name_src_ref
                )

            # Do type-check in array suffixes
            for array_suffix in array_suffixes:
                array_suffix.predict_type()

            # Check array suffixes
            if (isinstance(current_comp, comp.AddressableComponent)) and current_comp.is_array:
                assert isinstance(current_comp.array_dimensions, list)
                # is an array
                if len(array_suffixes) != len(current_comp.array_dimensions):
                    self.msg.fatal(
                        "Incompatible number of index dimensions after '%s'. Expected %d, found %d."
                        % (name, len(current_comp.array_dimensions), len(array_suffixes)),
                        name_src_ref
                    )
            elif array_suffixes:
                # Has array suffixes. Check if compatible with referenced component
                self.msg.fatal(
                    "Unable to index non-array component '%s'" % name,
                    name_src_ref
                )

        return type(current_comp)

    def get_value(self, eval_width: Optional[int]=None) -> rdltypes.ComponentRef:
        """
        Build a resolved ComponentRef container that describes the relative path
        """

        resolved_ref_elements = []

        for name, array_suffixes, name_src_ref in self.ref_elements:
            idx_list = [suffix.get_value() for suffix in array_suffixes]
            resolved_ref_elements.append((name, idx_list, name_src_ref))

        # Create container
        cref = rdltypes.ComponentRef(self.ref_root, resolved_ref_elements)

        return cref


class PropRef(ASTNode):
    def __init__(self, env: 'RDLEnvironment', src_ref: 'OptionalSourceRef', inst_ref: ASTNode, prop_ref_type: Type[rdltypes.PropertyReference]):
        super().__init__(env, src_ref)
        # InstRef to the component whose property is being referenced
        self.inst_ref = inst_ref

        # PropertyReference class
        self.prop_ref_type = prop_ref_type

    def predict_type(self) -> Type[rdltypes.PropertyReference]:
        """
        Predict the type of the inst_ref, and make sure the property being
        referenced is allowed
        """
        inst_type = self.inst_ref.predict_type()

        if self.prop_ref_type.allowed_inst_type != inst_type:
            self.msg.fatal(
                "'%s' is not a valid property of instance" % self.prop_ref_type.get_name(),
                self.src_ref
            )

        return self.prop_ref_type

    def get_value(self, eval_width: Optional[int]=None) -> rdltypes.PropertyReference:
        cref = self.inst_ref.get_value()
        return self.prop_ref_type(self.src_ref, self.env, cref)
